#pragma once
#include <iostream>
#include <cmath>
#include <vector>
#include <fstream>
#include <iomanip>

#define PI 3.14159265358979323846

template<class T>
class Point {
public:
    T x, y;
    Point(T x = 0, T y = 0) : x(x), y(y) {}
};

/// <summary>
/// Represents a geometric figure
/// </summary>
/// <typeparam name="T">Number value type</typeparam>
template<class T>
class GeometricFigure {
public:
    /// <summary>
    /// Calculate the area of the figure
    /// </summary>
    /// <returns>The internal area of the figure</returns>
    virtual T calc_area() const = 0;

    /// <summary>
    /// Calculate the perimeter of the figure
    /// </summary>
    /// <returns>The external perimeter of the figure</returns>
    virtual T calc_perimeter() const = 0;

    /// <summary>
    /// Get the name of the figure
    /// </summary>
    virtual void name() const = 0;

    friend std::ostream& operator<<(std::ostream& out, const GeometricFigure<T>& figure) {
        figure.name();
        out << " (perimeter: " << figure.calc_perimeter() << ", area: " << figure.calc_area() << ")";
        return out;
    }

    virtual ~GeometricFigure() = default;
};

/// <summary>
/// Represents a circle
/// </summary>
/// <typeparam name="T">Number value type</typeparam>
template<class T>
class Circle : public GeometricFigure<T> {
private:
    Point<T> center;
    T radius;

public:
    /// <summary>
    /// Create a circle by its center point and radius
    /// </summary>
    /// <param name="c">Center point of the circle</param>
    /// <param name="r">Radius of the circle</param>
    Circle(Point<T> c, T r) : center(c), radius(r) {}

    T calc_area() const override;
    T calc_perimeter() const override;
    void name() const override;
};

template<class T>
T Circle<T>::calc_area() const {
    return PI * radius * radius;
}

template<class T>
T Circle<T>::calc_perimeter() const {
    return 2 * PI * radius;
}

template<class T>
void Circle<T>::name() const {
    std::cout << "Circle";
}

/// <summary>
/// Represents an ellipse
/// </summary>
/// <typeparam name="T">Number value type</typeparam>
template<class T>
class Ellipse : public GeometricFigure<T> {
private:
    Point<T> center;
    T a, b;

public:
    /// <summary>
    /// Create an ellipse by its center point and two dimensions
    /// </summary>
    /// <param name="c">Center of the ellipse</param>
    /// <param name="a">First dimension</param>
    /// <param name="b">Second dimension</param>
    Ellipse(Point<T> c, T a, T b) : center(c), a(a), b(b) {}

    T calc_area() const override;
    T calc_perimeter() const override;
    void name() const override;
};

template<class T>
T Ellipse<T>::calc_area() const {
    return PI * a * b;
}

template<class T>
T Ellipse<T>::calc_perimeter() const {
    // Approximation of the perimeter of an ellipse
    return PI * (3 * (a + b) - std::sqrt((3 * a + b) * (a + 3 * b)));
}

template<class T>
void Ellipse<T>::name() const {
    std::cout << "Ellipse";
}

/// <summary>
/// Represents a triangle
/// </summary>
/// <typeparam name="T">Number value type</typeparam>
template<class T>
class Triangle : public GeometricFigure<T> {
private:
    Point<T> vertices[3];

    T side_length(const Point<T>& p1, const Point<T>& p2) const {
        return std::sqrt((p2.x - p1.x) * (p2.x - p1.x) + (p2.y - p1.y) * (p2.y - p1.y));
    }

public:
    /// <summary>
    /// Create a triangle by three points
    /// </summary>
    /// <param name="a">Point A</param>
    /// <param name="b">Point B</param>
    /// <param name="c">Point C</param>
    Triangle(Point<T> a, Point<T> b, Point<T> c) : vertices{ a, b, c } {}

    T calc_area() const override;
    T calc_perimeter() const override;
    void name() const override;
};

template<class T>
T Triangle<T>::calc_area() const {
    T a = side_length(vertices[0], vertices[1]);
    T b = side_length(vertices[1], vertices[2]);
    T c = side_length(vertices[2], vertices[0]);
    T s = (a + b + c) / 2;
    return std::sqrt(s * (s - a) * (s - b) * (s - c));
}

template<class T>
T Triangle<T>::calc_perimeter() const {
    T a = side_length(vertices[0], vertices[1]);
    T b = side_length(vertices[1], vertices[2]);
    T c = side_length(vertices[2], vertices[0]);
    return a + b + c;
}

template<class T>
void Triangle<T>::name() const {
    std::cout << "Triangle";
}

/// <summary>
/// Represents a rectangle
/// </summary>
/// <typeparam name="T">Number value type</typeparam>
template<class T>
class Rectangle : public GeometricFigure<T> {
private:
    Point<T> vertices[4];

    T side_length(const Point<T>& p1, const Point<T>& p2) const {
        return std::sqrt((p2.x - p1.x) * (p2.x - p1.x) + (p2.y - p1.y) * (p2.y - p1.y));
    }

public:
    /// <summary>
    /// Create a rectangle by 4 points
    /// </summary>
    /// <param name="a">Top-left point</param>
    /// <param name="b">Bottom-left point</param>
    /// <param name="c">Bottom-right point</param>
    /// <param name="d">Top-right point</param>
    Rectangle(Point<T> a, Point<T> b, Point<T> c, Point<T> d) : vertices{ a, b, c, d } {}

    T calc_area() const override;
    T calc_perimeter() const override;
    void name() const override;
};

template<class T>
T Rectangle<T>::calc_area() const {
    T width = side_length(vertices[0], vertices[1]);
    T height = side_length(vertices[1], vertices[2]);
    return width * height;
}

template<class T>
T Rectangle<T>::calc_perimeter() const {
    T width = side_length(vertices[0], vertices[1]);
    T height = side_length(vertices[1], vertices[2]);
    return 2 * (width + height);
}

template<class T>
void Rectangle<T>::name() const {
    std::cout << "Rectangle";
}

/// <summary>
/// Represents an abstract polygonal figure
/// </summary>
/// <typeparam name="T">Number value type</typeparam>
template<class T>
class Polygon : public GeometricFigure<T> {
private:
    std::vector<Point<T>> vertices;

public:
    /// <summary>
    /// Read polygon from file
    /// </summary>
    /// <param name="filename">Path to the file</param>
    Polygon(const std::string& filename);

    T calc_area() const override;
    T calc_perimeter() const override;
    void name() const override;
};

template<class T>
Polygon<T>::Polygon(const std::string& filename) {
    std::ifstream file(filename);
    if (!file.is_open()) {
        throw std::runtime_error("Error opening file");
    }
    int numVertices;
    file >> numVertices;
    for (int i = 0; i < numVertices; ++i) {
        T x, y;
        file >> x >> y;
        vertices.emplace_back(x, y);
    }
    file.close();
}

template<class T>
T Polygon<T>::calc_area() const {
    T area = 0.0;
    int n = vertices.size();
    for (int i = 0; i < n; ++i) {
        area += (vertices[i].x * vertices[(i + 1) % n].y) - (vertices[(i + 1) % n].x * vertices[i].y);
    }
    return 0.5 * std::abs(area);
}

template<class T>
T Polygon<T>::calc_perimeter() const {
    T perimeter = 0.0;
    int n = vertices.size();
    for (int i = 0; i < n; ++i) {
        perimeter += std::sqrt((vertices[(i + 1) % n].x - vertices[i].x) * (vertices[(i + 1) % n].x - vertices[i].x) + (vertices[(i + 1) % n].y - vertices[i].y) * (vertices[(i + 1) % n].y - vertices[i].y));
    }
    return perimeter;
}

template<class T>
void Polygon<T>::name() const {
    std::cout << "Polygon";
}
